#include "mpi.h"

typedef struct _table {
  int references;
  int length;
  int *value;
} Table;

int copy_table (oldcomm, newcomm, keyval, extra_state)
MPI_Comm oldcomm, newcomm;
int keyval;
void *extra_state;
{
  Table *table;
  int    found;

  MPI_Get_attr ( oldcomm, keyval, &table, &found );
  if (found) {
    table->references++;
    MPI_Put_attr ( newcomm, keyval, table );
  }
  (*(int *)extra_state)++;
  return 1;
}

void create_table ( num, values, table_out )
int  num;
int *values;
Table **table_out;
{
  int i;
  (*table_out) = (Table *)malloc(sizeof(Table));
  (*table_out)->references = 1;
  (*table_out)->length = num;
  (*table_out)->value = (int *)malloc(sizeof(int)*num);
  for (i=0;i<num;i++) 
    (*table_out)->value[i] = values[i];
}

void delete_table (comm, keyval, extra_state)
MPI_Comm comm;
int keyval;
void *extra_state;
{
  Table *table;
  int    found;

  MPI_Get_attr ( comm, keyval, &table, &found );
  if ( found ) {
    if ( table->references == 1 )
      free(table);
    else
      table->references--;
    (*(int *)extra_state)--;
  }
}

worker ( argc, argv )
int argc;
char **argv;
{
  int i, rank, size;
  Table *table;
  MPI_Comm new_comm;
  int table_key;
  int values[3]; 
  int table_copies = 1;
  int found;
  int errors = 0;

  MPI_Init ( &argc, argv );
  MPI_Comm_rank ( MPI_COMM_WORLD, &rank );
  MPI_Comm_size ( MPI_COMM_WORLD, &size );

  values[0] = 1; values[1] = 2; values[2] = 3;
  create_table(3,values,&table);

  MPI_Attr_create_key ( copy_table, delete_table, &table_key, &table_copies );
  MPI_Put_attr ( MPI_COMM_WORLD, table_key, table );
  MPI_Comm_dup ( MPI_COMM_WORLD, &new_comm );
  MPI_Get_attr ( new_comm, table_key, &table, &found );

  if ((table_copies != 2) && (table->references != 2)) 
    errors++;

  MPI_Comm_free ( &new_comm );

  if ((table_copies != 1) && (table->references != 1)) 
    errors++;

  MPI_Delete_attr ( MPI_COMM_WORLD, table_key );

  if ( table_copies != 0 )
    errors++;
  if (errors)
    printf("[%d] OOPS.  %d errors!\n",rank,errors);
  else
    printf("[%d] Successful\n",rank);

  MPI_Attr_free_key ( table_key );
  MPI_Finalize();
}






